import jmt.analytical.Solver;
import jmt.analytical.SolverMulti;
import jmt.analytical.SolverMultiMixed;
import jmt.analytical.SolverSingleClosedMVA;

public class PowerModelSolverAsync {
	
	static boolean log = false;
	
	public static double round(double d){
		return Math.round(d*100.0)/100.0;
		//return d;
	}

	public static boolean converged(int numDevices, double[] d, double[] nd,
			double[] r, double[] nr) {
		boolean converged = true;
		if(log) 
		{System.out.println("convergence test starting .. ");
		for(int i=0;i< numDevices; i++){
			System.out.println("d[" + i + "] = " + d[i] + " , r[" + i + "] = " + r[i]);
		}
		}
		for (int i = 0; i < numDevices; i++) {
			if (((d[i] - nd[i]) >= 0.05) || ((r[i] - nr[i]) >= 0.05)) {
				converged = false;
				break;
			}
		}
		//copy nd, nr into d,r
		for(int i=0;i<numDevices;i++){
			d[i] = nd[i];
			r[i] = nr[i];
		}
		return converged;
	}

	public static void main(String[] args) {

		if(args.length != 2){
			System.err.println("Error incorrect format ");
			System.exit(1);
		}
		
		int numDevices = Integer.parseInt(args[0]);
		int numCustomers = Integer.parseInt(args[1]);

		double delays[] = new double[numDevices];
		double newdelays[] = new double[numDevices];
		double servicetime[] = new double[numDevices];
		double newservicetime[] = new double[numDevices];
		double residencetimelambda[] = new double[numDevices];
		double probabilities[] = new double[numDevices];
		
		for (int i = 0; i < numDevices; i++) {
			newdelays[i] = newservicetime[i] = 0;
		}
		
		for (int i = 1; i < numDevices; i++) {
			servicetime[i] = 0.04;
		}
		
		servicetime[0] = 0.01;
		
		//  probability inputs
		for(int i = 0; i < numDevices; i++){
			probabilities[i] = 1.00;
		}
		
		for (int i = 0; i < numDevices; i++) {
			delays[i] = 2;
		}
		
		// primary task network
		SolverSingleClosedMVA primary = null;
		String[] pname = { "CPU", "Delay" };
		int[] ptype = { Solver.LI, Solver.DELAY };
		double[] pvisit = { 1, 1 };
		double[][] pservt = new double[2][1];

		// secondary task network
		SolverMultiMixed[] secondaries = new SolverMultiMixed[numDevices];
		String[] sname = new String[3];
		int[] stype = new int[3];
		int[] classTypes = new int[2];
		double[] classData = new double[2];
		double[][] svisit = new double[3][2]; // service_station, class
		double[][][] sservt = new double[3][2][1]; // service_station,class
		// names
		sname[0] = "Device";
		sname[1] = "SibDelay";
		sname[2] = "ReentryDelay";
		// type of servers
		stype[0] = Solver.LI;
		stype[1] = Solver.DELAY;
		stype[2] = Solver.DELAY;
		// class types
		classTypes[0] = SolverMulti.OPEN_CLASS;
		classTypes[1] = SolverMulti.CLOSED_CLASS;
		// class data - arrival rate if open class else num_customers
		classData[1] = numCustomers;

		// service center 1
		svisit[0][0] = 1;
		svisit[0][1] = 1;
		// service center 2
		svisit[1][0] = 0;
		svisit[1][1] = 1;
		// service center 3
		svisit[2][0] = 0;
		svisit[2][1] = 1;
		
		int iterationCount = 1;

		do{
			
			if(log) System.out.println(iterationCount + " iteration starting ...");
			/* STEP 0 - Solve primary with given delay[0] & servicetime[0]
			 * to get residence time[0]
			 */
			
			primary = new SolverSingleClosedMVA(numCustomers, 2);
			pservt[0][0] = servicetime[0];
			pservt[1][0] = delays[0];

			if (!primary.input(pname, ptype, pservt, pvisit)) {
				System.err.println("Primary task network input failed");
				System.exit(1);
			} else {
				primary.solve();
			}
			
			double primaryResidenceTime = primary.getResTime(0);
			
			if(log) System.out.println(iterationCount + " Primary Residence time " + primaryResidenceTime);

			/* STEP 1 - Solve the product form networks for secondary tasks */
			// new nr[i] is created for all secondary tasks;
			for (int i = 1; i < numDevices; i++) {
				secondaries[i] = new SolverMultiMixed(2, 3);

				// right now constant rate of asynchronous tasks for all
				// secondary servers
				classData[0] = 0.25;

				// class 0 - open customers
				sservt[1][0][0] = 0;
				sservt[2][0][0] = 0;
				// class 1 - closed customers
				sservt[1][1][0] = delays[i];
				sservt[2][1][0] = ((primaryResidenceTime + delays[0])/probabilities[i]) - delays[0];
				// both class 0 & 1 at Device have same service time
				sservt[0][0][0] = sservt[0][1][0] = servicetime[i];

				if (!secondaries[i].input(sname, stype, sservt, svisit,classData, classTypes)) {
					System.err.println("Secondary task network input failed");
					System.exit(1);
				} else {
					secondaries[i].solve();
				}
				
				if (secondaries[i].getResTime(0, 0) != secondaries[i]
						.getResTime(0, 1)) {
					/*System.err.println("Residence times did not match ..."
							+ secondaries[i].getResTime(0, 0) + " "
							+ secondaries[i].getResTime(0, 1));*/
				}
				newservicetime[i] = 1/secondaries[i].getThroughput(0, 1);
				residencetimelambda[i] = 1/secondaries[i].getResTime(0, 1);
				
			}
			
			if(log){
			for(int i=1;i<numDevices;i++){
				System.out.println(iterationCount + " Secondary " + i + " Res time rate " + residencetimelambda[i] + " " + 1/residencetimelambda[i] + " service time " + newservicetime[i] );
			}
			}

			/*
			 * STEP 2 - Calculate the wait time of primary task for all
			 * secondary tasks it created
			 */
			// new nd[0] is created
			
			newdelays[0] = 0;
			int[] indices;
			for (int i = 1; i < numDevices; i++) {
				double termsum = 0;
				CombinationGenerator comb = new CombinationGenerator(
						numDevices-1, i);
				while (comb.hasMore()) {
					indices = comb.getNext();
					boolean valid = true;
					if(log) System.out.print("Combination :");
					for (int j = 0; j < indices.length-1; j++) {
						if (indices[j] >= indices[j + 1]) {
							valid = false;
							break;
						}
					}
					if(log){
					for(int j=0;j<indices.length;j++){
						System.out.print(indices[j]);
					}
					System.out.println(" valid ? " + valid);
					}
					if (valid) {
						double sum = 0;
						for (int j = 0; j < indices.length; j++) {
							sum += residencetimelambda[indices[j]+1];
						}
						termsum += 1 / sum;
					}
				}
				if(log) System.out.println(iterationCount + " newdelays[0] = " + newdelays[0] + " termsum " + termsum );
				newdelays[0] += java.lang.Math.pow(-1,i-1) * termsum; 
			}
			if(log)
				System.out.println(iterationCount + " Delay at primary " + newdelays[0]);

			/* STEP 3 - Calculate new nd[i] for all secondary tasks */
			for (int i = 1; i < numDevices; i++) {
				newdelays[i] = newdelays[0] - (1/residencetimelambda[i]);
				if(log) System.out.println(iterationCount + " Delay at secondary " + i + " is " + newdelays[i]);
			}

			/* STEP 4 - Solve the product form network for primary task */
			// new nr[0] is created for the primary task
			primary = new SolverSingleClosedMVA(numCustomers, 2);
			pservt[0][0] = servicetime[0];
			pservt[1][0] = newdelays[0];

			if (!primary.input(pname, ptype, pservt, pvisit)) {
				System.err.println("Primary task network input failed");
				System.exit(1);
			} else {
				primary.solve();
			}

			newservicetime[0] = 1/primary.getThroughput(0);
			residencetimelambda[0] =  1/primary.getResTime(0);
			
			if(log) System.out.println(iterationCount + " Primary task service time " + newservicetime[0] + " residence time rate " + residencetimelambda[0]);
			iterationCount++;
		}while (!converged(numDevices, delays, newdelays, servicetime, newservicetime));
		
/*		//System.out.println("\n\nService times\n");
		for(int i=0;i<numDevices;i++){
			System.out.print(newservicetime[i]);
			if(i != numDevices -1){
				System.out.print(",");
			}
		}*/
		
		//System.out.println("\n\nQueue lengths\n");
		System.out.print(round(primary.getQueueLen(0)) + " ");
		for (int i = 1; i < numDevices; i++) {
			System.out.print(round(secondaries[i].getQueueLen(0,1)));
			if (i != numDevices - 1) {
				System.out.print(" ");
			}else{
				System.out.println(" ");
			}
		}
		
		//System.out.println("\n\nUtilization\n");
		System.out.print(round(primary.getUtilization(0)) + " ");
		for (int i = 1; i < numDevices; i++) {
			System.out.print( round(secondaries[i].getUtilization(0,1)));
			if (i != numDevices - 1) {
				System.out.print(" ");
			}else{
				System.out.println(" ");
			}
		}
		
		//System.out.println("\n\nThroughput\n");
		System.out.print(round(primary.getThroughput(0)) + " ");
		for (int i = 1; i < numDevices; i++) {
			System.out.print( round(secondaries[i].getThroughput(0,1)) );
			if (i != numDevices - 1) {
				System.out.print(" ");
			}else{
				System.out.println(" ");
			}
		}
		
	}
}