import java.text.DecimalFormat;

import jmt.analytical.Solver;
import jmt.analytical.SolverSingleClosedMVA;

public class PowerModelSolverOrig {

	static boolean log = false;
	static double convergenceThreshold = 0.05;
	
	public static double round(double d){
		return Math.round(d*100.0)/100.0;
		//return d;
	}

	public static boolean converged(int numDevices, double[] d, double[] nd,
			double[] r, double[] nr) {
		boolean converged = true;
		
		for (int i = 0; i < numDevices; i++) {
			if ((Math.abs(d[i] - nd[i]) >= convergenceThreshold) || (Math.abs(r[i] - nr[i]) >= convergenceThreshold)) {
				converged = false;
				break;
			}
		}
		if (log) {
			System.out.println("convergence test : " + converged);
			for (int i = 0; i < numDevices; i++) {
				System.out.println("d[" + i + "] = " + d[i] + " , nd[" + i
						+ "] = " + nd[i] + "r[" + i + "] = " + r[i] + " , nr[" + i
						+ "] = " + nr[i]);
			}
		}
		// copy nd, nr into d,r
		for (int i = 0; i < numDevices; i++) {
			d[i] = nd[i];
			r[i] = nr[i];
		}
		return converged;
	}

	public static void main(String[] args) {

		if(args.length != 2){
			System.err.println("Error incorrect format ");
			System.exit(1);
		}
		
		int numDevices = Integer.parseInt(args[0]);
		int numCustomers = Integer.parseInt(args[1]);
		
		//System.out.println("numDevices " + numDevices + " numCustomers " + numCustomers);

		double delays[] = new double[numDevices];
		double newdelays[] = new double[numDevices];
		double servicetime[] = new double[numDevices];
		double newservicetime[] = new double[numDevices];
		double residencetimelambda[] = new double[numDevices];
		double probabilities[] = new double[numDevices];

		for (int i = 0; i < numDevices; i++) {
			newdelays[i] = newservicetime[i] = 0;
		}

		for (int i = 1; i < numDevices; i++) {
			servicetime[i] = 0.04;
		}

		servicetime[0] = 0.01;

		// probability inputs
		for (int i = 0; i < numDevices; i++) {
			probabilities[i] = 1.00;
		}

		for (int i = 0; i < numDevices; i++) {
			delays[i] = 2;
		}

		// primary task network
		SolverSingleClosedMVA primary = null;
		String[] pname = { "CPU", "Delay" };
		int[] ptype = { Solver.LI, Solver.DELAY };
		double[] pvisit = { 1, 1 };
		double[][] pservt = new double[2][1];

		// secondary task network
		SolverSingleClosedMVA[] secondaries = new SolverSingleClosedMVA[numDevices];
		String[] sname = { "Devince", "SibDelay", "ReentryDelay" };
		int[] stype = { Solver.LI, Solver.DELAY, Solver.DELAY };
		double[] svisit = { 1, 1, 1 };
		double[][] sservt = new double[3][1];

		int iterationCount = 1;

		do {

			if (log)
				System.out.println(iterationCount + " iteration starting ...");
			/*
			 * STEP 0 - Solve primary with given delay[0] & servicetime[0] to
			 * get residence time[0]
			 */

			primary = new SolverSingleClosedMVA(numCustomers, 2);
			pservt[0][0] = servicetime[0];
			pservt[1][0] = delays[0];

			if (!primary.input(pname, ptype, pservt, pvisit)) {
				System.err.println("Primary task network input failed");
				System.exit(1);
			} else {
				primary.solve();
			}

			double primaryResidenceTime = primary.getResTime(0);

			if (log)
				System.out.println(iterationCount + " Primary Residence time "
						+ primaryResidenceTime);

			/* STEP 1 - Solve the product form networks for secondary tasks */
			// new nr[i] is created for all secondary tasks;
			for (int i = 1; i < numDevices; i++) {
				secondaries[i] = new SolverSingleClosedMVA(numCustomers, 3);

				sservt[1][0] = delays[i];
				sservt[2][0] = ((primaryResidenceTime + delays[0]) / probabilities[i])
						- delays[0];
				sservt[0][0] = servicetime[i];

				if (!secondaries[i].input(sname, stype, sservt, svisit)) {
					System.err.println("Secondary task network input failed");
					System.exit(1);
				} else {
					secondaries[i].solve();
				}
				newservicetime[i] = 1 / secondaries[i].getThroughput(0);
				residencetimelambda[i] = 1 / secondaries[i].getResTime(0);

			}

			if (log) {
				for (int i = 1; i < numDevices; i++) {
					System.out.println(iterationCount + " Secondary " + i
							+ " Res time rate " + residencetimelambda[i] + " "
							+ 1 / residencetimelambda[i] + " service time "
							+ newservicetime[i]);
				}
			}

			/*
			 * STEP 2 - Calculate the wait time of primary task for all
			 * secondary tasks it created
			 */
			// new nd[0] is created

			newdelays[0] = 0;
			int[] indices;
			for (int i = 1; i < numDevices; i++) {
				double termsum = 0;
				CombinationGenerator comb = new CombinationGenerator(
						numDevices - 1, i);
				while (comb.hasMore()) {
					indices = comb.getNext();
					boolean valid = true;
					if (log)
						System.out.print("Combination :");
					for (int j = 0; j < indices.length - 1; j++) {
						if (indices[j] >= indices[j + 1]) {
							valid = false;
							break;
						}
					}
					if (log) {
						for (int j = 0; j < indices.length; j++) {
							System.out.print(indices[j]);
						}
						System.out.println(" valid ? " + valid);
					}
					if (valid) {
						double sum = 0;
						for (int j = 0; j < indices.length; j++) {
							sum += residencetimelambda[indices[j] + 1];
						}
						termsum += 1 / sum;
					}
				}
				if (log)
					System.out.println(iterationCount + " newdelays[0] = "
							+ newdelays[0] + " termsum " + termsum);
				newdelays[0] += java.lang.Math.pow(-1, i - 1) * termsum;
			}
			if (log)
				System.out.println(iterationCount + " Delay at primary "
						+ newdelays[0]);

			/* STEP 3 - Calculate new nd[i] for all secondary tasks */
			for (int i = 1; i < numDevices; i++) {
				newdelays[i] = newdelays[0] - (1 / residencetimelambda[i]);
				if (log)
					System.out.println(iterationCount + " Delay at secondary "
							+ i + " is " + newdelays[i]);
			}

			/* STEP 4 - Solve the product form network for primary task */
			// new nr[0] is created for the primary task
			primary = new SolverSingleClosedMVA(numCustomers, 2);
			pservt[0][0] = servicetime[0];
			pservt[1][0] = newdelays[0];

			if (!primary.input(pname, ptype, pservt, pvisit)) {
				System.err.println("Primary task network input failed");
				System.exit(1);
			} else {
				primary.solve();
			}

			newservicetime[0] = 1 / primary.getThroughput(0);
			residencetimelambda[0] = 1 / primary.getResTime(0);

			if (log)
				System.out.println(iterationCount
						+ " Primary task throughput " + primary.getThroughput(0) 
						+ " Primary task service time " + newservicetime[0]
						+ " residence time rate " + residencetimelambda[0]);
			iterationCount++;
		} while (!converged(numDevices, delays, newdelays, servicetime,
				newservicetime));

		//System.out.println("\n\nService times\n");
		for (int i = 0; i < numDevices; i++) {
			System.out.print(round(newservicetime[i]));
			if (i != numDevices - 1) {
				System.out.print(" ");
			}else{
				System.out.println(" ");
			}
		}

		//System.out.println("\n\nResidence times\n");
		for (int i = 0; i < numDevices; i++) {
			System.out.print(round((1 / residencetimelambda[i])));
			if (i != numDevices - 1) {
				System.out.print(" ");
			}else{
				System.out.println(" ");
			}
		}

		//System.out.println("\n\nDelay times\n");
		for (int i = 0; i < numDevices; i++) {
			System.out.print(round(newdelays[i]));
			if (i != numDevices - 1) {
				System.out.print(" ");
			}else{
				System.out.println(" ");
			}
		}

		//System.out.println("\n\nQueue lengths\n");
		System.out.print(round(primary.getQueueLen(0)) + " ");
		for (int i = 1; i < numDevices; i++) {
			System.out.print(round(secondaries[i].getQueueLen(0)));
			if (i != numDevices - 1) {
				System.out.print(" ");
			}else{
				System.out.println(" ");
			}
		}
		
		//System.out.println("\n\nUtilization\n");
		System.out.print(round(primary.getUtilization(0)) + " ");
		for (int i = 1; i < numDevices; i++) {
			System.out.print( round(secondaries[i].getUtilization(0)));
			if (i != numDevices - 1) {
				System.out.print(" ");
			}else{
				System.out.println(" ");
			}
		}
		
		//System.out.println("\n\nThroughput\n");
		System.out.print(round(primary.getThroughput(0)) + " ");
		for (int i = 1; i < numDevices; i++) {
			System.out.print( round(secondaries[i].getThroughput(0)) );
			if (i != numDevices - 1) {
				System.out.print(" ");
			}else{
				System.out.println(" ");
			}
		}
	}
}